from sphinx_subfigure import __version__

project = "Sphinx Subfigures"
author = "Chris Sewell"
copyright = "2022, Chris Sewell"
version = __version__

extensions = ["myst_parser", "sphinx_subfigure"]

myst_enable_extensions = ["colon_fence"]
numfig = True

suppress_warnings = ["epub.unknown_project_files"]

html_theme = "furo"
html_static_path = ["_static"]
html_css_files = ["custom.css"]



from __future__ import annotations

from html import escape
from typing import Any

from docutils import nodes
from sphinx.application import Sphinx
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.writers.html import HTMLTranslator

from ._compat import findall


def setup_html(app: Sphinx):
    """Setup the extension for HTML building."""
    app.add_post_transform(SubfigureHtmlTransform)
    app.add_node(
        SubfigureGridHtml, html=(visit_subfigure_grid_html, depart_subfigure_grid_html)
    )
    app.add_node(
        SubfigureGridItemHtml,
        html=(visit_subfigure_grid_item_html, depart_subfigure_grid_item_html),
    )
    app.connect("html-page-context", html_page_context)


def html_page_context(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: dict[str, Any],
    doctree: nodes.document | None,
) -> None:
    """Add CSS for grid layouts, to HTML pages that require it."""
    if not doctree:
        return
    if "subfig_layouts" not in doctree:
        return

    style = []
    for size, cls_layout in doctree["subfig_layouts"].items():
        if size == "sm":
            style.append("@media (max-width: 576px) {")
        if size == "lg":
            style.append("@media (min-width: 992px) {")
        if size == "xl":
            style.append("@media (min-width: 1200px) {")
        if size == "xxl":
            style.append("@media (min-width: 1400px) {")
        for layout_class, layout in cls_layout.items():
            layout_grid = " ".join(["'" + " ".join(row) + "'" for row in layout])
            style.append(f"  .{layout_class} {{ grid-template-areas: {layout_grid}; }}")
        if size in ["sm", "lg", "xl", "xxl"]:
            style.append("}")

    if style:
        context["body"] = (
            "\n<style>\n" + "\n".join(style) + "\n</style>\n" + context["body"]
        )


class SubfigureGridHtml(nodes.General, nodes.Element):
    """Node for subfigure grid, added only for Html."""


def visit_subfigure_grid_html(self: HTMLTranslator, node: SubfigureGridHtml) -> None:
    """Visit subfigure grid node."""
    classes = " ".join(node["classes"])
    style = "display: grid;"
    if "gap" in node:
        # the grid- prefix is deprecated
        style += f" gap: {node['gap']}; grid-gap: {node['gap']};"
    self.body.append(f'<div class="{classes}" style="{style}">\n')


def depart_subfigure_grid_html(self: HTMLTranslator, node: SubfigureGridHtml) -> None:
    """Depart subfigure grid node."""
    self.body.append("</div>\n")


class SubfigureGridItemHtml(nodes.General, nodes.Element):
    """Node for subfigure grid item, added only for Html."""


def visit_subfigure_grid_item_html(
    self: HTMLTranslator, node: SubfigureGridItemHtml
) -> None:
    """Visit subfigure grid item node."""
    classes = " ".join(node["classes"])
    style = "display: flex; flex-direction: column; justify-content: center; align-items: center;"
    style += f" grid-area: {node['area']};"
    self.body.append(f'<div class="{classes}" style="{style}">\n')
    if "caption" in node and node["caption-align"] == "above":
        self.body.append(f'<span class="caption">{escape(node["caption"])}</span>\n')


def depart_subfigure_grid_item_html(
    self: HTMLTranslator, node: SubfigureGridItemHtml
) -> None:
    """Depart subfigure grid item node."""
    if "caption" in node and node["caption-align"] == "below":
        self.body.append(f'<span class="caption">{escape(node["caption"])}</span>\n')
    self.body.append("</div>\n")


class SubfigureHtmlTransform(SphinxPostTransform):
    """Transform subfigure containers into the HTML specific AST structures."""

    default_priority = 199
    formats = ("html",)

    def run(self) -> None:
        """Run the transform."""

        # docutils <0.18 (traverse) >=0.18 (findall) compatibility
        for fig_node in findall(
            self.document, lambda n: "is_subfigure" in getattr(n, "attributes", {})
        ):

            # initiate figure children
            children = []

            # store layouts
            classes = ["sphinx-subfigure-grid"]
            if "subfig_layouts" not in self.document:
                self.document["subfig_layouts"] = {}
            for size, layout in fig_node["layout"].items():
                layout_class = f"ss-layout-{size}-" + "_".join(
                    ["".join(a.replace(".", "d") for a in row) for row in layout]
                )
                self.document["subfig_layouts"].setdefault(size, {})[
                    layout_class
                ] = layout
                classes.append(layout_class)

            # add grid
            grid_node = SubfigureGridHtml(
                classes=classes + fig_node.get("grid_classes", [])
            )
            if "gap" in fig_node:
                grid_node["gap"] = fig_node["gap"]
            children.append(grid_node)

            # add image items to grid
            caption = None
            area_classes = ["sphinx-subfigure-area"] + fig_node.get("area_classes", [])
            for child in fig_node:
                if isinstance(child, nodes.caption):
                    caption = child
                    continue
                item_node = SubfigureGridItemHtml(classes=area_classes)
                # if fig_node["layout_type"] == "areas":
                item_node["area"] = child["subfigure_area"]
                if fig_node.get("subcaptions") and child.get("alt"):
                    item_node["caption-align"] = fig_node["subcaptions"]
                    item_node["caption"] = child["alt"]
                item_node.append(child)
                grid_node.append(item_node)

            # add caption
            if caption:
                children.append(caption)

            fig_node.children = children



import base64
from pathlib import Path
import re

from bs4 import BeautifulSoup
import pytest
from sphinx_pytest.plugin import CreateDoctree

IMAGE_PNG = base64.b64decode(
    "iVBORw0KGgoAAAANSUhEUgAAAHQAAAB0CAYAAABUmhYnAAAEd0lEQVR4Xu2c0ZLjIAwEk///6GzVvZlspWtWksNRnVcwiGmNwHaS5+v1ej38HKPAU6DHsPy3EIGexVOgh/EUqEBPU+Cw9biHCvQwBQ5bjg4V6GEKHLYcHSrQwxQ4bDk6VKCHKXDYcnSoQA9T4LDllB36fD5vlWR9fUvz0+ve9fp0/O7FU7w0n0CXhBSoDiXTRO06FBKKBLLkLvlGgkTp+UvndPzu/ul46Xq7x2/fQ8kR0wtOBaL+1J6uZ+3fPb5Aw0PRtxOWEkigAr3mCJUMuk9cM45uG3ZvJwel8dN4byW8+r1cgWYPVgRaLIlpwqWCT1cgHbr8skOgYUqkgtHwVYfQKZTiTW8rdCgQFWjtt2Pjty3TGdztOB0aHlosuVcHpglJ+h3nUFow7bE6dDOHCjRN2fBty917qEAF+jEHaI+bTlhK0Nsf/aUBpXtYdXy6noDS9dTePf74oYgWRO3dC6b57k6o7vUJFAh3Cz6dMAIV6FWB9FCQlry1f/ejQXLgt9eX6tXu0DSAtL9APysm0OYHI2mCUgVKxxOoQNOcubc/7XnF5yj3LuYPs5Ud+oc5Ry8R6GEpK1CBjlaMuwcvl1xyBC2I8im9T0xva6pPbtL1V+MjPQW6KEQJRAlAggs0vK2oCibQ4g9+LbnXb96THlQBvl5y0yclqYNQAKgAVGIJQHWPpfjf4uv+bUsagECvClCCkL46VIdecyQtKZRhlKGW3OG3LekeQ0DSBOk+1VLCdbdTAqfzlUuuQFPJe/fM9kORQAV6UYBKJslF11NJS0s8xZO2U3zpeO0lNw2g2+HV8dLbKJov1aMKWKDFfyITKKRsegqmjE7H06FpTRHoRwUoQUnu9pJLh4z0EFMdjwRI46ESWwVC8VK7QMN/TRHookDqCB1Knry261AdmmXMdG86xabzd49H83fP1+5QWkB3e7sg4eu06nra46++4K4uqHp9uyACrSKpXS/Q5kMRnUJruN6vnr7Po/VMn9KrepX3UBKgGmD1UVw6P61HoKmi0F+HfhZIhy766NDhU2F66CEgzQXjQRUjjb8aX7tDaYFpwKkgAi0SSAUXaO0Pjkk/HUoKFQ9p0wm/hjcONC2B6W3B24KKv1ZLx0vzgfQoFsyHQJe3LQINHUEZrUNre6wO1aHLw+AvO5QOHdReLbE0/vSeedyhKBWUDh00XpoAAg2/EkIAqD0FlPYXqEDp3Pix/b8/FKUOIMem7fR6j8Yr0fvlYoEWK4JAw0dplOE6dLnrqH5JrCp4NcMFejPQ6h7RnTAUT/eTKkpYiidtH99D04C6bwvS+QX65W8sUMkVaKgAlcRwuLfuNL5Ah/fQKkC6Pi2JKXB6NEjxUTslKF1P7e17KE1YbRfoZwUFuuijQ4v/l5s6VocOOzQFYv9ZBcoldzY8R08VEGiq2Ob9Bbo5oDQ8gaaKbd5foJsDSsMTaKrY5v0FujmgNDyBpopt3l+gmwNKwxNoqtjm/QW6OaA0PIGmim3eX6CbA0rDE2iq2Ob9Bbo5oDS8H8eCMw7yCzx+AAAAAElFTkSuQmCC"  # noqa: E501
)

FIXTURE_PATH = Path(__file__).parent.joinpath("fixtures")


@pytest.mark.param_file(FIXTURE_PATH / "posttransform_html.txt")
def test_posttransform_html(file_params, sphinx_doctree: CreateDoctree):
    """Test AST output after post-transforms, when using the HTML builder."""
    sphinx_doctree.set_conf({"extensions": ["sphinx_subfigure"]})
    sphinx_doctree.buildername = "html"
    sphinx_doctree.srcdir.joinpath("image.png").write_bytes(IMAGE_PNG)
    result = sphinx_doctree(file_params.content)
    assert not result.warnings
    file_params.assert_expected(result.get_resolved_pformat(), rstrip_lines=True)


@pytest.mark.param_file(FIXTURE_PATH / "build_html.txt")
def test_build_html(file_params, sphinx_doctree: CreateDoctree):
    """Test HTML build output."""
    sphinx_doctree.set_conf({"extensions": ["sphinx_subfigure"]})
    sphinx_doctree.buildername = "html"
    sphinx_doctree.srcdir.joinpath("image.png").write_bytes(IMAGE_PNG)
    result = sphinx_doctree(file_params.content)
    assert not result.warnings
    html = BeautifulSoup(
        Path(result.builder.outdir).joinpath("index.html").read_text(), "html.parser"
    )
    fig_html = str(html.select_one("figure.sphinx-subfigure"))
    file_params.assert_expected(fig_html, rstrip_lines=True)


@pytest.mark.param_file(FIXTURE_PATH / "posttransform_latex.txt")
def test_posttransform_latex(file_params, sphinx_doctree: CreateDoctree):
    """Test AST output after post-transforms, when using the LaTeX builder."""
    sphinx_doctree.set_conf({"extensions": ["sphinx_subfigure"]})
    sphinx_doctree.buildername = "latex"
    sphinx_doctree.srcdir.joinpath("image.png").write_bytes(IMAGE_PNG)
    result = sphinx_doctree(file_params.content)
    assert not result.warnings
    file_params.assert_expected(result.get_resolved_pformat(), rstrip_lines=True)


@pytest.mark.param_file(FIXTURE_PATH / "build_latex.txt")
def test_build_latex(file_params, sphinx_doctree: CreateDoctree):
    """Test LaTeX build output."""
    sphinx_doctree.set_conf({"extensions": ["sphinx_subfigure"]})
    sphinx_doctree.buildername = "latex"
    sphinx_doctree.srcdir.joinpath("image.png").write_bytes(IMAGE_PNG)
    result = sphinx_doctree(file_params.content)
    assert not result.warnings
    tex = Path(result.builder.outdir).joinpath("python.tex").read_text()
    fig_tex = re.findall(r"\\begin\{figure\}.*\\end\{figure\}", tex, re.DOTALL)[0]
    file_params.assert_expected(fig_tex, rstrip_lines=True)


@pytest.mark.param_file(FIXTURE_PATH / "posttransform_man.txt")
def test_posttransform_man(file_params, sphinx_doctree: CreateDoctree):
    """Test AST output after post-transforms, when using the Man builder."""
    sphinx_doctree.set_conf({"extensions": ["sphinx_subfigure"]})
    sphinx_doctree.buildername = "man"
    sphinx_doctree.srcdir.joinpath("image.png").write_bytes(IMAGE_PNG)
    result = sphinx_doctree(file_params.content)
    assert not result.warnings
    file_params.assert_expected(result.get_resolved_pformat(), rstrip_lines=True)


@pytest.mark.param_file(FIXTURE_PATH / "build_html.txt")
def test_build_man(file_params, sphinx_doctree: CreateDoctree):
    """Test Man build."""
    sphinx_doctree.set_conf({"extensions": ["sphinx_subfigure"]})
    sphinx_doctree.buildername = "man"
    sphinx_doctree.srcdir.joinpath("image.png").write_bytes(IMAGE_PNG)
    result = sphinx_doctree(file_params.content)
    assert not result.warnings



from __future__ import annotations

import string

from docutils import nodes
from sphinx.application import Sphinx
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.writers.latex import LaTeXTranslator

from ._compat import findall


def setup_latex(app: Sphinx):
    """Setup the extension for LaTeX building."""
    app.add_latex_package("subcaption")
    app.add_post_transform(SubfigureLaTexTransform)
    app.add_node(
        SubfigureEnvLatex,
        latex=(visit_subfigure_grid_latex, depart_subfigure_grid_latex),
    )


class SubfigureEnvLatex(nodes.General, nodes.Element):
    """Node for subfigure grid, added only for LaTeX."""


def visit_subfigure_grid_latex(self: LaTeXTranslator, node: SubfigureEnvLatex) -> None:
    """Visit subfigure grid node."""
    self.body.append("\\begin{subfigure}{" + str(node["width"]) + "\\textwidth}\n")
    self.body.append("\\centering\n")
    if "caption" in node and node["caption-align"] == "above":
        self.body.append("\\caption{" + self.escape(node["caption"]) + "}\n")


def depart_subfigure_grid_latex(self: LaTeXTranslator, node: SubfigureEnvLatex) -> None:
    """Depart subfigure grid node."""
    if "caption" in node and node["caption-align"] == "below":
        self.body.append("\\caption{" + self.escape(node["caption"]) + "}\n")
    self.body.append("\\end{subfigure}\n")
    if node.get("new-row"):
        self.body.append("\n")


class SubfigureLaTexTransform(SphinxPostTransform):
    """Transform subfigure containers into the HTML specific AST structures."""

    default_priority = 199
    formats = ("latex",)

    def run(self) -> None:
        """Run the transform."""

        # docutils <0.18 (traverse) >=0.18 (findall) compatibility
        for fig_node in findall(
            self.document, lambda n: "is_subfigure" in getattr(n, "attributes", {})
        ):
            layout = fig_node["layout"]["default"]
            if not layout:
                continue
            # if the layout is a simple progression of areas,
            # we can use just the subfigure environment to create a grid
            progression = True
            next = (0, "A")
            flattened = []
            for row in layout:
                for idx, area in enumerate(row):
                    if area == ".":
                        pass
                    elif idx != 0 and flattened[-1][0] == area:
                        flattened[-1][1] += 1
                    elif area == next[1]:
                        flattened.append([area, 1, False])
                        next = (next[0] + 1, string.ascii_uppercase[next[0] + 1])
                    else:
                        progression = False
                        break
                if not progression:
                    break
                flattened[-1][2] = True

            flattened[-1][2] = False

            if not progression:
                continue

            width = round(0.99 / len(layout[0]), 3)

            children = []
            subfigs = 0
            for child in fig_node:
                if isinstance(child, nodes.caption):
                    children.append(child)
                else:
                    subcaption_node = SubfigureEnvLatex(
                        width=width * flattened[subfigs][1]
                    )
                    if flattened[subfigs][2]:
                        subcaption_node["new-row"] = True
                    subfigs += 1
                    if fig_node.get("subcaptions") and child.get("alt"):
                        subcaption_node["caption-align"] = fig_node["subcaptions"]
                        subcaption_node["caption"] = child["alt"]
                    subcaption_node += child
                    children.append(subcaption_node)
            fig_node.children = children



from __future__ import annotations

import math
import string

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective

from .tr_html import setup_html
from .tr_latex import setup_latex


def setup(app: Sphinx) -> None:
    """Setup the extension."""
    app.add_directive("subfigure", SubfigureDirective)
    setup_html(app)
    setup_latex(app)


class SubfigureDirective(SphinxDirective):
    """A sphinx directive to create sub-figures."""

    has_content = True
    required_arguments = 0
    optional_arguments = 1  # the subfigure layout (defaults to 1)
    final_argument_whitespace = True
    option_spec = {
        "layout-sm": directives.unchanged,
        "layout-lg": directives.unchanged,
        "layout-xl": directives.unchanged,
        "layout-xxl": directives.unchanged,
        "subcaptions": lambda x: directives.choice(x, ("above", "below")),
        "width": directives.length_or_percentage_or_unitless,
        "align": lambda x: directives.choice(x, ("left", "center", "right")),
        "gap": directives.length_or_percentage_or_unitless,
        "name": directives.unchanged,
        "class": directives.class_option,
        "class-grid": directives.class_option,
        "class-area": directives.class_option,
    }

    def run(self) -> list[nodes.Element]:
        """Run the directive."""
        self.assert_has_content()
        figure_node = nodes.figure(
            is_subfigure=True,
            classes=["sphinx-subfigure"] + self.options.get("class", []),
            grid_classes=self.options.get("class-grid", []),
            area_classes=self.options.get("class-area", []),
        )
        self.set_source_info(figure_node)
        for attribute_name in ("align", "width", "gap", "subcaptions"):
            if attribute_name in self.options:
                figure_node.attributes[attribute_name] = self.options[attribute_name]
        self.add_name(figure_node)
        self.state.nested_parse(self.content, self.content_offset, figure_node)

        number_of_images = 0
        has_caption = False
        for idx, child in enumerate(list(figure_node)):
            if isinstance(child, nodes.image):
                child["subfigure_area"] = string.ascii_uppercase[number_of_images]
                number_of_images += 1
            elif (
                isinstance(child, nodes.paragraph)
                and child.children
                and isinstance(child[0], nodes.image)
            ):
                images = []
                for sub in child:
                    if not isinstance(sub, nodes.image):
                        continue
                    images.append(sub)
                    sub["subfigure_area"] = string.ascii_uppercase[number_of_images]
                    number_of_images += 1
                child.replace_self(images)
            elif isinstance(child, nodes.paragraph):
                if has_caption:
                    raise self.error("Invalid subfigure content (multiple captions)")
                child.replace_self(nodes.caption(child.rawsource, *child.children))
                has_caption = True
            else:
                raise self.error(
                    "Invalid subfigure content; must contain only images and single caption, "
                    f"item {idx + 1} is neither (line {child.line})"
                )

        print(number_of_images)
        layout_string = self.arguments[0] if self.arguments else 1
        figure_node["layout"] = {}
        figure_node["layout"]["default"] = self.generate_layout(
            layout_string, number_of_images
        )
        for size in ("sm", "lg", "xl", "xxl"):
            layout_opt = f"layout-{size}"
            if self.options.get(layout_opt):
                figure_node["layout"][size] = self.generate_layout(
                    self.options.get(layout_opt), number_of_images, ltype=layout_opt
                )

        return [figure_node]

    def generate_layout(
        self, string: str, items: int, ltype: str = "layout", validate: bool = True
    ) -> list[list[str]]:
        """Generate a layout from a string."""
        layout = self._generate_layout(string, items)
        if validate:
            self._validate_layout(layout, items, ltype)
        return layout

    def _validate_layout(
        self, layout: list[list[str]], items: int, ltype="layout"
    ) -> None:
        """Validate a layout.

        Validated according to https://www.w3.org/TR/css3-grid-layout/#grid-template-areas-property:
        - All areas must be named A-Z.
        - All areas must form a single filled-in rectangle
        """

        prefix = f"Invalid subfigure {ltype}"

        area_indices: dict[str, list[tuple[int, int]]] = {}

        # check all rows have the same number of columns, and retrieve indices of each area
        if not layout:
            raise self.error(f"{prefix} (empty)")
        for rid, row in enumerate(layout):
            if not row:
                raise self.error(f"{prefix} (empty row)")
            if len(row) != len(layout[0]):
                raise self.error(f"{prefix} (row length mismatch)")
            for cid, area in enumerate(row):
                if area != ".":
                    area_indices.setdefault(area, set()).add((rid, cid))

        available_areas = {string.ascii_uppercase[i] for i in range(items)}
        used_areas = set(area_indices)

        # check all areas are used in the layout
        missing_areas = available_areas - used_areas
        if missing_areas:
            raise self.error(f"{prefix} (missing areas {missing_areas})")

        # check all areas correspond to an item
        additional_areas = used_areas - available_areas
        if additional_areas:
            raise self.error(f"{prefix} (invalid areas {additional_areas})")

        # Check that all area_indices form single filled-in rectangles
        for area, indices in area_indices.items():
            if len(indices) == 1:
                continue
            x0, y0 = min(x for x, _ in indices), min(y for _, y in indices)
            x1, y1 = max(x for x, _ in indices), max(y for _, y in indices)
            expected_indices = {
                (x, y) for y in range(y0, y1 + 1) for x in range(x0, x1 + 1)
            }
            if expected_indices != indices:
                raise self.error(f"{prefix} (area {area} is not a single rectangular)")

    def _generate_layout(self, layout_string: str, items: int) -> list[list[str]]:
        """Generate a figure layout from a string."""
        layout = []

        # if an integer is given, generate a layout with this many columns
        # this is similar to the CSS: grid-template-columns: repeat(columns, 1fr)
        try:
            layout_columns = int(layout_string)
        except ValueError:
            pass
        else:
            for row in range(math.ceil(items / layout_columns)):
                layout.append([])
                for col in range(layout_columns):
                    if (row * layout_columns) + col > items - 1:
                        area = "."
                    else:
                        area = string.ascii_uppercase[(row * layout_columns) + col]
                    layout[row].append(area)
            return layout

        # if a string is given, parse it as a grid layout, with columns delimited by "|",
        # ignore spaces, named areas A-Z and empty areas are filled with "."
        for row_string in layout_string.split("|"):
            row = []
            for col in row_string:
                if not col:
                    continue  # ignore spaces
                row.append(col)
            layout.append(row)
        return layout
